@@ -20,26 +20,25 @@
 #define __PointCutExpr_h__
 
 #include "Puma/Array.h"
+#include "Puma/Location.h"
 using namespace Puma;
 
 #include "Condition.h"
 #include "PointCut.h"
-#include "MatchExpr.h"
-#ifdef ACMODEL
-#include "Elements.h"
-#else
-#include "JoinPointLoc.h"
-#endif
 
-class PointCutContext;
+#include "ACModel/MatchExpr.h"
+#include "ACModel/Elements.h"
+#include "ACModel/Utils.h"
+
+#include <list>
+#include <string>
+
 class Binding;
+class PointCutContext;
 
 namespace Puma {
-  class CFunctionInfo;
-  class CArgumentInfo;
   class CTypeInfo;
   class ErrorStream;
-  class CTree;
 } // namespace Puma
 
 // Pointcut expression types are evaluated in a context-independant manner.
@@ -52,36 +51,38 @@ enum PCE_Type {
 
 class PointCutExpr {
 
-  CTree *_node;
+  int _ref_count;
 
 protected:
 
-#ifdef ACMODEL
+  enum { PRE_EVAL_OFF, PRE_EVAL_ON, PRE_EVAL_DONE } _pre_evaluate;
+  ACM_Any *_jpl;
   JoinPointType _possible_types;
-#else
-  JoinPointLoc::join_point_type _possible_types;
-#endif
 
   // do a semantic analysis of the child nodes
   void sem_args (ErrorStream &err, PointCutContext &context);
 
   // check if the argument pointcut types are as expected
-  void check_arg_types (ErrorStream &err, const char *func, PCE_Type expected);
+  void check_arg_types (ErrorStream &err, const char *func, PCE_Type expected,
+      PointCutContext &context);
 
   // check if all argument types are equal
-  void check_arg_types_equal (ErrorStream &err, const char *func);
+  void check_arg_types_equal (ErrorStream &err, const char *func,
+      PointCutContext &context);
 
   // get the real argument binding of a context variable
-  const CArgumentInfo *get_binding (const char *name, ErrorStream &err, 
-				    PointCutContext &context);
+  const ACM_Arg *get_binding (const char *name, ErrorStream &err,
+      PointCutContext &context);
 public:
 
-  PointCutExpr () : _node (0) {}
+  PointCutExpr () : _ref_count (1), _pre_evaluate (PRE_EVAL_OFF) {}
   virtual ~PointCutExpr () {}
 
+  // get/modify the reference count
+  int get_ref () const { return _ref_count; }
+  void ref(int delta = 1) { _ref_count += delta; }
+
   // set and get the corresponding syntax tree and expression child nodes
-  void node (CTree *node) { _node = node; }
-  CTree *node () const { return _node; }
   virtual int args () const = 0;
   virtual PointCutExpr *arg (int) const = 0;
 
@@ -97,16 +98,40 @@ public:
 
   // evaluate an expression, i.e. check if the expression evaluates as 'true'
   // for a particular join point
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond) = 0;
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond) = 0;
-#endif
+
+  // if the expression supports pre-evaluation, do it in this function
+  virtual void pre_evaluate (PointCutContext &context) {}
+
+  bool match (ACM_Any &jpl, PointCutContext &context,
+              Binding &binding, Condition &cond) {
+    if (_pre_evaluate == PRE_EVAL_ON)
+      pre_evaluate(context);
+    if (_pre_evaluate == PRE_EVAL_OFF)
+      return evaluate (jpl, context, binding, cond);
+    return (&jpl == _jpl);
+  }
                          
   // mangle a string that represents a type check (for that/target arguments)                         
-  virtual void mangle_type_check (ostream &) = 0;                         
+  virtual void mangle_type_check (ostream &) = 0;
+
+  // duplicate the whole pointcut expression tree -- at least logically
+  virtual PointCutExpr* duplicate () = 0;
+
+  // destroy a pointcut expression tree
+  // TODO: not a really intuitive solution, find a better one ...
+  static void destroy (PointCutExpr * expr) {
+    if (!expr) return;
+    expr->ref(-1);
+    if (expr->get_ref () == 0) {
+      for (int i = 0; i < expr->args (); i++)
+        destroy (expr->arg (i));
+      delete expr;
+    }
+  }
+
+
 };
 
 class PCE_SingleArg : public PointCutExpr {
@@ -116,8 +141,8 @@ public:
 
   virtual int args () const { return 1; }
   virtual PointCutExpr *arg (int i) const { return (i == 0) ? _arg : 0; }
+  void replace_arg (PointCutExpr *expr) { if (_arg) _arg->ref(-1); _arg = expr; }
 
-#ifdef ACMODEL
   bool check_derived_class (ACM_Class &cls,
     PointCutContext &context, Binding &binding, Condition &cond);
   bool check_derived_func (ACM_Class &cls, ACM_Function &func, 
@@ -128,18 +153,6 @@ public:
     PointCutContext &context, Binding &binding, Condition &cond);
   bool check_scopes (ACM_Name *scope,
     PointCutContext &context, Binding &binding, Condition &cond);
-#else
-  bool check_derived_class (JPL_Class &cls,
-    PointCutContext &context, Binding &binding, Condition &cond);
-  bool check_derived_func (JPL_Class &cls, JPL_Function &func, 
-    PointCutContext &context, Binding &binding, Condition &cond);
-  bool check_base_class (JPL_Class &cls, PointCutContext &context,
-    Binding &binding, Condition &cond);
-  bool check_base_func (JPL_Class &cls, JPL_Function &func, 
-    PointCutContext &context, Binding &binding, Condition &cond);
-  bool check_scopes (JPL_Name *scope,
-    PointCutContext &context, Binding &binding, Condition &cond);
-#endif
   
   virtual void mangle_type_check (ostream &out) {
     out << 1;
@@ -183,17 +196,13 @@ public:
   PCE_Classes (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "classes_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Classes* duplicate () { return new PCE_Classes(arg(0)->duplicate ()); }
 };
 
 class PCE_Base : public PCE_SingleArg {
@@ -201,17 +210,13 @@ public:
   PCE_Base (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "base_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Base* duplicate () { return new PCE_Base(arg(0)->duplicate ()); }
 };
 
 class PCE_Derived : public PCE_SingleArg {
@@ -219,17 +224,13 @@ public:
   PCE_Derived (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "derived_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Derived* duplicate () { return new PCE_Derived(arg(0)->duplicate ()); }
 };
 
 class PCE_Within : public PCE_SingleArg {
@@ -237,17 +238,13 @@ public:
   PCE_Within (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "within_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Within* duplicate () { return new PCE_Within(arg(0)->duplicate ()); }
 };
 
 class PCE_Execution : public PCE_SingleArg {
@@ -255,17 +252,13 @@ public:
   PCE_Execution (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "execution_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Execution* duplicate () { return new PCE_Execution(arg(0)->duplicate ()); }
 };
 
 class PCE_Call : public PCE_SingleArg {
@@ -273,17 +266,13 @@ public:
   PCE_Call (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "call_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Call* duplicate () { return new PCE_Call(arg(0)->duplicate ()); }
 };
 
 class PCE_Construction : public PCE_SingleArg {
@@ -291,17 +280,13 @@ public:
   PCE_Construction (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "construction_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Construction* duplicate () { return new PCE_Construction(arg(0)->duplicate ()); }
 };
 
 class PCE_Destruction : public PCE_SingleArg {
@@ -309,17 +294,13 @@ public:
   PCE_Destruction (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "destruction_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Destruction* duplicate () { return new PCE_Destruction(arg(0)->duplicate ()); }
 };
 
 class PCE_That : public PCE_SingleArg {
@@ -327,17 +308,13 @@ public:
   PCE_That (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "that_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_That* duplicate () { return new PCE_That(arg(0)->duplicate ()); }
 };
 
 class PCE_Target : public PCE_SingleArg {
@@ -345,17 +322,13 @@ public:
   PCE_Target (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "target_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Target* duplicate () { return new PCE_Target(arg(0)->duplicate ()); }
 };
 
 class PCE_CFlow : public PCE_SingleArg {
@@ -367,17 +340,13 @@ public:
   virtual PCE_Type type () const;
   PointCut &arg_pointcut () { return _arg_pointcut; }
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "cflow_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_CFlow* duplicate () { return new PCE_CFlow(arg(0)->duplicate ()); }
 };
 
 class PCE_Args : public PCE_VarArgs {
@@ -385,17 +354,18 @@ public:
   virtual ~PCE_Args () {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "args_";
     PCE_VarArgs::mangle_type_check (out);
   }
+  virtual PCE_Args* duplicate () {
+    PCE_Args *result = new PCE_Args;
+    for (int i = 0; i < args(); i++)
+      result->add_arg(arg(i)->duplicate ());
+    return result;
+  }
 };
 
 class PCE_Result : public PCE_SingleArg {
@@ -403,17 +373,13 @@ public:
   PCE_Result (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "result_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Result* duplicate () { return new PCE_Result(arg(0)->duplicate ()); }
 };
 
 class PCE_Or : public PCE_DoubleArg {
@@ -422,17 +388,13 @@ public:
     PCE_DoubleArg (arg0, arg1) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "or_";
     PCE_DoubleArg::mangle_type_check (out);
   }
+  virtual PCE_Or* duplicate () { return new PCE_Or(arg(0)->duplicate (), arg(1)->duplicate()); }
 };
 
 class PCE_And : public PCE_DoubleArg {
@@ -441,17 +403,13 @@ public:
     PCE_DoubleArg (arg0, arg1) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "and_";
     PCE_DoubleArg::mangle_type_check (out);
   }
+  virtual PCE_And* duplicate () { return new PCE_And(arg(0)->duplicate (), arg(1)->duplicate()); }
 };
 
 class PCE_Not : public PCE_SingleArg {
@@ -459,83 +417,73 @@ public:
   PCE_Not (PointCutExpr *arg) : PCE_SingleArg (arg) {}
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "not_";
     PCE_SingleArg::mangle_type_check (out);
   }
+  virtual PCE_Not* duplicate () { return new PCE_Not(arg(0)->duplicate ()); }
 };
 
 class PCE_Named : public PCE_SingleArg {
-  CFunctionInfo *_func;
+  ACM_Pointcut *_pct_func;
+  std::list<std::string> _params;
 public:
-  PCE_Named (CFunctionInfo *func, PointCutExpr *expr) : 
-    PCE_SingleArg (expr), _func (func) {}
+  PCE_Named (ACM_Pointcut *pct_func, PointCutExpr *expr) :
+    PCE_SingleArg (expr), _pct_func (pct_func) {}
   virtual PCE_Type type () const;
+  std::list<std::string> &params () { return _params; }
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out) {
     out << "named_";
-    PCE_SingleArg::mangle_type_check (out);
+    arg(0)->mangle_type_check (out);
+  }
+  virtual PCE_Named* duplicate () {
+    PCE_Named *result = new PCE_Named(_pct_func, arg (0) ? arg(0)->duplicate() : 0);
+    result->_params = _params;
+    return result;
   }
 };
 
 class PCE_ContextVar : public PointCutExpr {
-  const char *_name;
-  const CArgumentInfo *_bound_arg;
+  string _name;
+  const ACM_Arg *_bound_arg;
   const CTypeInfo *_check_type;
   enum { CV_THAT, CV_TARGET, CV_ARG, CV_RESULT } _bound_to;
   int _arg;
 public:
-  PCE_ContextVar (const char *name) : _name (name) {}
+  PCE_ContextVar (const string &name) : _name (name) {}
   virtual int args () const { return 0; }
   virtual PointCutExpr *arg (int) const { return 0; }
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out);
+  virtual PCE_ContextVar* duplicate () { return new PCE_ContextVar(_name); }
 };
 
 class PCE_Match : public PointCutExpr {
-#ifdef ACMODEL
-  ACM::MatchExpr _match_expr;
-#else
   MatchExpr _match_expr;
-#endif
   string _str;
 public:
-  PCE_Match (const char *str) : _str (str) {}
+  PCE_Match (const string &str) : _str (str) {}
   virtual ~PCE_Match () {}
   virtual int args () const { return 0; }
   virtual PointCutExpr *arg (int) const { return 0; }
   virtual PCE_Type type () const;
   virtual void semantics (ErrorStream &err, PointCutContext &context);
-#ifdef ACMODEL
+
+  // if the expression supports pre-evaluation, do it in this function
+  virtual void pre_evaluate (PointCutContext &context);
+  void pre_evaluate (const MatchName &name, PointCutContext &context);
   virtual bool evaluate (ACM_Any &jpl, PointCutContext &context,
                          Binding &binding, Condition &cond);
-#else
-  virtual bool evaluate (JoinPointLoc &jpl, PointCutContext &context,
-                         Binding &binding, Condition &cond);
-#endif
   virtual void mangle_type_check (ostream &out);
+  virtual PCE_Match* duplicate () { ref(1); return this; }
 };
 
 #endif // __PointCutExpr_h__
