@@ -16,35 +16,19 @@
 // Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 // MA  02111-1307  USA                                            
 
-#include <assert.h>
-
 #include "ThisJoinPoint.h"
-#include "JoinPoint.h"
 #include "Naming.h"
 #include "Binding.h"
-#include "JPAdvice.h"
-#ifdef ACMODEL
-#include "Utils.h"
-#else
-#include "JoinPointLoc.h"
-#endif
-#include "WeaverBase.h"
-
-#include "Puma/CTree.h"
-#include "Puma/CClassInfo.h"
-#include "Puma/CFunctionInfo.h"
-#include "Puma/CTypeInfo.h"
-#include "Puma/CArgumentInfo.h"
+#include "ACModel/Utils.h"
+#include "ResultBuffer.h"
+#include "AdviceInfo.h"
 
 #include <sstream>
 using std::stringstream;
 using std::endl;
-#include <string>
-using std::string;
 
 // special setup function for bindings
 void ThisJoinPoint::setup (const Binding &binding) {
-  _setup = true;
   if (binding._used) {
     _used |= (binding._this != 0) ? THAT : 0;
     _used |= (binding._target != 0) ? TARGET : 0;
@@ -54,111 +38,64 @@ void ThisJoinPoint::setup (const Binding &binding) {
   }
 }
 
-void ThisJoinPoint::setup (CFunctionInfo* func) {
-  assert (func->Tree ()->NodeName () == CT_FctDef::NodeId ());
-
-  _setup = true;
-  // by starting with the second statement of the body we avoid to iterate
-  // through the generated argument list and the generated typedef at the
-  // beginning of the function
-  CT_CmpdStmt *body = ((CT_FctDef*)func->Tree ())->Body ();
-//  for (int i = 1; i < body->Entries (); i++)
-// EXPERIMENT
-  for (int i = 0; i < body->Entries (); i++)
-    setup_search (func, body->Entry (i));
+void ThisJoinPoint::check_field (const char* field, bool dyn_only) {
+  if (strcmp (field, "signature") == 0)
+    _used |= SIGNATURE;
+  else if (strcmp (field, "filename") == 0)
+    _used |= FILENAME;
+  else if (strcmp (field, "line") == 0)
+    _used |= LINE;
+  else if (strcmp (field, "args") == 0)
+    _used |= ARGS;
+  else if (strcmp (field, "arg") == 0)
+    _used |= ARG;
+  else if (strcmp (field, "argtype") == 0)
+    _used |= ARG_TYPE;
+  else if (strcmp (field, "type") == 0)
+    _used |= TYPE;
+  else if (strcmp (field, "id") == 0 || (!dyn_only && strcmp (field, "JPID") == 0))
+    _used |= ID;
+  else if (strcmp (field, "resulttype") == 0)
+    _used |= RESULT_TYPE;
+  else if (strcmp (field, "that") == 0)
+    _used |= THAT;
+  else if (strcmp (field, "target") == 0)
+    _used |= TARGET;
+  else if (strcmp (field, "result") == 0)
+    _used |= RESULT;
+  else if (strcmp (field, "jptype") == 0)
+    _used |= JP_TYPE;
+  else if (strcmp (field, "action") == 0)
+    _used |= ACTION;
+  else if (strcmp (field, "wrapper") == 0)
+    _used |= WRAPPER;
+  else if (strcmp (field, "proceed") == 0)
+    _used |= PROCEED;
 }
 
-void ThisJoinPoint::setup_search (CFunctionInfo *func, CTree *node) {
-  CObjectInfo *obj;
-  CRecord *cls;
-
-  const char *nodename = node->NodeName ();
-  
-  if (nodename == CT_SimpleName::NodeId () &&
-      (obj = ((CT_SimpleName*)node)->Object ())) {
-
-    if (obj->Scope ()->GlobalScope () &&
-        strcmp (obj->Name (), "tjp") == 0)
-      _used |= (PTR_NEEDED | TYPE_NEEDED);
-    else if (obj->Scope ()->GlobalScope () &&
-             strcmp (obj->Name (), "thisJoinPoint") == 0)
-      _used |= (PTR_NEEDED | PTR_ALIAS_NEEDED | TYPE_NEEDED);
-    else if (obj->Scope ()->GlobalScope () &&
-             strcmp (obj->Name (), "JoinPoint") == 0)
-      _used |= TYPE_NEEDED;
-  
-    if ((cls = obj->ClassScope ()) &&
-        strcmp (cls->QualName (), "JoinPoint") == 0) {
-      const char *field = obj->Name ();
-
-      if (strcmp (field, "signature") == 0)
-        _used |= SIGNATURE;
-      else if (strcmp (field, "filename") == 0)
-        _used |= FILENAME;
-      else if (strcmp (field, "line") == 0)
-        _used |= LINE;
-      else if (strcmp (field, "args") == 0)
-        _used |= ARGS;
-      else if (strcmp (field, "arg") == 0)
-        _used |= ARG;
-      else if (strcmp (field, "argtype") == 0)
-        _used |= ARG_TYPE;
-      else if (strcmp (field, "type") == 0)
-        _used |= TYPE;
-      else if (strcmp (field, "id") == 0)
-        _used |= ID;
-      else if (strcmp (field, "resulttype") == 0)
-        _used |= RESULT_TYPE;
-      else if (strcmp (field, "that") == 0)
-        _used |= THAT;
-      else if (strcmp (field, "target") == 0)
-        _used |= TARGET;
-      else if (strcmp (field, "result") == 0)
-        _used |= RESULT;
-      else if (strcmp (field, "jptype") == 0)
-        _used |= JP_TYPE;
-      else if (strcmp (field, "action") == 0)
-        _used |= ACTION;
-      else if (strcmp (field, "wrapper") == 0)
-        _used |= WRAPPER;
-      else if (strcmp (field, "proceed") == 0)
-        _used |= PROCEED;
-    }
-
+bool ThisJoinPoint::check_type (const string &name) {
+  if (name == "JoinPoint") {
+    _used |= TYPE_NEEDED;
+    return true;
   }
+  return false;
+}
 
-  for (int s = 0; s < node->Sons (); s++)
-    setup_search (func, node->Son (s));
-}   
-
-#ifdef ACMODEL
-string ThisJoinPoint::gen_result_type (ACM_Code *jpl) {
-  CTypeInfo *rtype = (has_result_type(*jpl)?
-                      TI_Type::of (*get_result_type(*jpl))->type_info ():
-                      &CTYPE_UNDEFINED);
-#else
-string ThisJoinPoint::gen_result_type (JPL_Code *jpl) {
-  CTypeInfo *rtype = TI_Type::of (jpl->result_type ())->type_info ();
-#endif
-  if (rtype->isVoid () || rtype->isUndefined ())
-    return "void";
-  else if (rtype->isAddress ())
-    rtype = rtype->BaseType ();
-  stringstream type;
-  rtype->TypeText (type, "", true, true);
-  return type.str ();
+bool ThisJoinPoint::check_obj (const string &name) {
+  if (name == "tjp") {
+    _used |= (PTR_NEEDED | TYPE_NEEDED);
+    return true;
+  }
+  else if (name == "thisJoinPoint") {
+    _used |= (PTR_NEEDED | PTR_ALIAS_NEEDED | TYPE_NEEDED);
+    return true;
+  }
+  return false;
 }
 
-#ifdef ACMODEL
 void ThisJoinPoint::gen_tjp_struct (ostream &code, ACM_Code *loc,
                                     BackEndProblems &bep, int depth) const {
-#else
-void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
-                                    BackEndProblems &bep, int depth) const {
-#endif
   
-  TI_Code *ti = (TI_Code*)loc->transform_info ();
-
   // the level 0 struct -> contains all the data and types
   code << "template <typename TResult, typename TThat, typename TTarget, ";
   code << "typename TArgs> struct ";
@@ -182,16 +119,16 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
   stringstream types, data, fct;
 
   // start: type definitions --------------------------------------------------
-#ifdef ACMODEL
-  CTypeInfo *rtype  = (has_result_type(*loc) ?
-                       TI_Type::of (*get_result_type (*loc))->type_info ():
-                       &CTYPE_UNDEFINED);
-#else
-  CTypeInfo *rtype  = TI_Type::of (loc->result_type ())->type_info ();
-#endif
-  bool res_is_ref   = rtype->isAddress ();
-  bool res_is_undef = rtype->isUndefined ();
-
+  bool res_is_undef = !has_result_type(*loc);
+  bool res_is_ref   = false;
+  if (!res_is_undef) {
+    ACM_Type *rtype = get_result_type (*loc);
+    MatchSignature &match_sig = rtype->get_match_sig();
+    if (match_sig.is_new())
+      match_sig.parse(format_type (*rtype));
+    if (match_sig.type ().is_reference())
+      res_is_ref = true;
+  }
   types << "  typedef TResult " << (res_is_ref ? "*" : "") << "Result;" << endl;
   types << "  typedef TThat   That;" << endl;
   types << "  typedef TTarget Target;" << endl;
@@ -201,18 +138,13 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
   types << "  template <int I> struct Arg : AC::Arg<TArgs, I> {};" << endl;
     
   // join point id and type;
-#ifdef ACMODEL
   // TODO: handle join-point ID in TJP
-//  int jpid = (loc->assigned_id () == -1 ? loc->id () : loc->assigned_id ());
-  types << "  static const int JPID = " << 0 /*jpid*/ << ";" << endl;
+  if (id()) {
+    int jpid = loc->get_jpid(); //(loc->assigned_id () == -1 ? loc->id () : loc->assigned_id ());
+    types << "  static const int JPID = " << jpid << ";" << endl;
+  }
   types << "  static const AC::JPType JPTYPE = (AC::JPType)" << loc->type_val ()
         << ";" << endl;
-#else
-  int jpid = (loc->assigned_id () == -1 ? loc->id () : loc->assigned_id ());
-  types << "  static const int JPID = " << jpid << ";" << endl;
-  types << "  static const AC::JPType JPTYPE = (AC::JPType)" << loc->type ()
-        << ";" << endl;
-#endif
   
   // result type
   types << "  struct Res {" << endl;
@@ -222,38 +154,20 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
   types << "  };" << endl;
   
   // argument types
-#ifdef ACMODEL
   unsigned arg_count = get_arg_count (*loc);
-#else
-  unsigned arg_count = loc->arg_count ();
-#endif
   
   // begin of static data -----------------------------------------------------
   if (signature ()) {
     fct << "  inline static const char *signature () { return \"";
-#ifdef ACMODEL
     fct << ::signature (*loc) << "\";}" << endl;
-#else
-    fct << loc->signature () << "\";}" << endl;
-#endif
   }
   if (filename ()) {
     fct << "  inline static const char *filename () { return \"";
-#ifdef ACMODEL
-    // TODO: implement filename function
     fct << ::filename (*loc) << "\";}" << endl;
-#else
-    fct << loc->filename () << "\";}" << endl;
-#endif
   }
   if (line ()) {
     fct << "  inline static int line () { return ";
-#ifdef ACMODEL
-    // TODO: implement line function
     fct << ::line(*loc) << ";}" << endl;
-#else
-    fct << loc->line () << ";}" << endl;
-#endif
   }
   if (args ()) {
     fct << "  inline static const int args () { return ";
@@ -262,12 +176,15 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
   }
   if (type()) {
     fct << "  inline static  AC::Type type() { return ";
-    if (ti->tjp_type()) {
-      fct << "\"";
-      ti->tjp_type()->TypeInfo()->Mangled(fct);
-      fct << "\";";
-    } else fct << "\"<unknown type>\";";
-    fct << "}" << endl;
+    fct << "\"";
+    if (has_result_type(*loc))
+      Naming::mangle (fct, get_result_type(*loc));
+    list<ACM_Type*> tlist;
+    get_arg_types (*loc, tlist);
+    for (list<ACM_Type*>::iterator iter = tlist.begin(); iter != tlist.end();
+        ++iter)
+      Naming::mangle (fct, *iter);
+    fct << "\"; }" << endl;
   }
   if (id()) {
     fct << "  inline static unsigned int id() { return JPID; }" << endl;
@@ -275,15 +192,9 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
   
   if (resulttype()) {
     fct << "  inline static AC::Type resulttype() {return ";
-#ifdef ACMODEL
     if (has_result_type(*loc)) {
       fct << "\"";
-      TI_Type::of (*get_result_type(*loc))->type_info()->Mangled(fct);
-#else
-      if (!TI_Type::of (loc->result_type ())->type_info ()->isUndefined ()) {
-        fct << "\"";
-        TI_Type::of (loc->result_type())->type_info()->Mangled(fct);
-#endif
+      Naming::mangle (fct, get_result_type(*loc));
       fct << "\";";
     } else fct << "\"<unknown signature>\";";
     fct << "}" << endl;    
@@ -297,7 +208,6 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
     fct << "  inline static AC::Type const argtype(unsigned i) {" << endl;
     if (arg_count > 0) {
       fct << "    static AC::Type const type[" << arg_count << "] = {";
-#ifdef ACMODEL
       list<ACM_Type*> tlist;
       get_arg_types (*loc, tlist);
       unsigned i = 0;
@@ -305,17 +215,9 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
            ++iter, ++i) {
         if (i > 0) fct << ", ";
         fct << "\"";
-        TI_Type::of (**iter)->type_info ()->Mangled(fct);
+        Naming::mangle (fct, *iter);
         fct << "\"";
       }
-#else
-      for (unsigned i = 0; i < arg_count; i++) {
-        if (i > 0) fct << ", ";
-        fct << "\"";
-        TI_Type::of (loc->arg_type (i))->type_info ()->Mangled(fct);
-        fct << "\"";
-      }
-#endif
       fct << "};" << endl;
       fct << "    return type[i];" << endl;
       fct << "  }" << endl;
@@ -326,11 +228,7 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
   }
    
   // begin of dynamic data ----------------------------------------------------
-#ifdef ACMODEL
   if (arg () || useAction () || (proceed () && proceed_needs_args (*loc))) {
-#else
-    if (arg () || useAction () || (proceed () && loc->proceed_needs_args ())) {
-#endif
     if (!useAction () && arg_count > 0)
       data << "  void *_args[ARGS];" << endl;
     fct <<  "  inline void *arg (int n) {return "
@@ -341,32 +239,24 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
     fct <<  "  }" << endl;
   }
 
-#ifdef ACMODEL
   if (result() || useAction () || (proceed () && proceed_needs_result (*loc))) {
-#else
-  if (result() || useAction () || (proceed () && loc->proceed_needs_result ())) {
-#endif
     if (!useAction ())
       data << "  Result *_result;" << endl;
     fct <<  "  inline Result *result() {return (Result*)_result;}" << endl;
   }
 
-  if (target() || useAction () || (proceed () && ti->proceed_needs_target ())) {
+  if (target() || useAction () || (proceed () && proceed_needs_target (*loc))) {
     if (!useAction ())
       data << "  Target *_target;" << endl;
     fct <<  "  inline Target *target() {return (Target*)_target;}" << endl;
   }
 
-  if (that() || useAction () || (proceed () && ti->proceed_needs_that ())) {
+  if (that() || useAction () || (proceed () && proceed_needs_that (*loc))) {
     if (!useAction ())
       data << "  That *_that;" << endl;
     fct <<  "  inline That *that() {return (That*)_that;}" << endl;
   }
 
-//  if (proceed () && ti->proceed_needs_fptr ()) {
-//    data << "  void *_fptr;" << endl;
-//  }
-  
   if (wrapper() || useAction ()) {
     if (!useAction ())
       data << "  void (*_wrapper)(AC::Action &);" << endl;
@@ -388,20 +278,24 @@ void ThisJoinPoint::gen_tjp_struct (ostream &code, JPL_Code *loc,
   // the closing bracket is *NOT* generated here -> for external extensions
 }
 
-#ifdef ACMODEL
+ACM_Function *ThisJoinPoint::target_func (ACM_Code *loc) {
+  ACM_Function *result = 0;
+  if (loc->type_val() == JPT_Call)
+    result = ((ACM_Call*)loc)->get_target();
+  else if (loc->type_val() & (JPT_Execution|JPT_Construction|JPT_Destruction))
+    result = (ACM_Function*)loc->get_parent();
+  return result;
+}
+
+ACM_Function *ThisJoinPoint::that_func (ACM_Code *loc) {
+  return (((ACM_Any*)loc->get_parent ())->type_val() == JPT_Function) ?
+      (ACM_Function*)loc->get_parent () : (ACM_Function*)0;
+}
+
 void ThisJoinPoint::gen_tjp_init (ostream &code, ACM_Code *loc,
-                                  BackEndProblems &bep, int depth) const {
-#else
-void ThisJoinPoint::gen_tjp_init (ostream &code, JPL_Code *loc,
-                                    BackEndProblems &bep, int depth) const {
-#endif
-  TI_Code *ti = (TI_Code*)TransformInfo::of (*loc);
+                                  BackEndProblems &bep, int depth, bool is_dep, vector<string> *arg_names) const {
   const ThisJoinPoint &tjp = *this; // TODO: not necessary!
-#ifdef ACMODEL
   JoinPointType jptype = loc->type_val ();
-#else
-  JoinPointLoc::join_point_type jptype = loc->type ();
-#endif
   
   if (tjp.pointer_needed ()) {    
 
@@ -409,32 +303,27 @@ void ThisJoinPoint::gen_tjp_init (ostream &code, JPL_Code *loc,
     Naming::tjp_instance(code, loc);
     code << ";" << endl;
 
-#ifdef ACMODEL
     int args = get_arg_count(*loc);
-#else
-    int args = ((JPL_Code*)loc)->arg_count ();
-#endif
-    if (tjp.arg() || tjp.useAction() ||
-#ifdef ACMODEL
-      (tjp.proceed () && proceed_needs_args (*loc))) {
-#else
-      (tjp.proceed () && loc->proceed_needs_args ())) {
-#endif
-        if (args) {
+    if (arg_needed (loc)) {
+      if (args) {
       	if (tjp.useAction ()) {
 	        code << "  void *";
   	      Naming::tjp_args_array(code, loc);
     	    code << "[] = { ";
       	  for (int i = 0; i < args; i++) {
         	  if (i > 0) code << ", ";
-          	code << "(void*)&arg" << i;
+          	code << "(void*)&";
+          	if (arg_names)
+          	  code << (*arg_names)[i];
+          	else
+          	  code << "arg" << i;
         	}
         	code << " };" << endl;
       	}
       }
 
-	if (tjp.useAction ()) {
-	    code << "  ";
+	    if (tjp.useAction ()) {
+	      code << "  ";
   	    Naming::tjp_instance(code, loc);
     	  code << "._args = ";
       	if (args)
@@ -447,66 +336,51 @@ void ThisJoinPoint::gen_tjp_init (ostream &code, JPL_Code *loc,
     	  for (int i = 0; i < args; i++) {
 		      code << "  ";
   		    Naming::tjp_instance(code, loc);
-    		  code << "._args[" << i << "] = (void*)&arg" << i << ";" << endl;
+    		  code << "._args[" << i << "] = (void*)&";
+          if (arg_names)
+            code << (*arg_names)[i];
+          else
+            code << "arg" << i;
+          code << ";" << endl;
     	  }
 			}
     }
     
     if (tjp.result() || tjp.useAction() ||
-#ifdef ACMODEL
       (tjp.proceed () && proceed_needs_result (*loc))) {
-#else
-      (tjp.proceed () && loc->proceed_needs_result ())) {
-#endif
         code << "  ";
       Naming::tjp_instance(code, loc);
       code << "._result = ";
-#ifdef ACMODEL
-      CTypeInfo *rtype = (has_result_type(*loc) ?
-                          TI_Type::of (*get_result_type(*loc))->type_info ():
-                          &CTYPE_UNDEFINED);
-#else
-      CTypeInfo *rtype = TI_Type::of (((JPL_Code*)loc)->result_type ())->type_info ();
-#endif
-      if (!(rtype->isVoid() || rtype->isUndefined ())) {
+      if (!has_result_type(*loc) || get_result_type(*loc)->get_signature() == "void") {
+        code << "0";
+      } else {
         if (tjp.useAction ())
           code << "(void*)";
-        code << "&(__TJP::Result&)result";
-      } else {
-        code << "0";
+        code << "&(" << (is_dep ? "typename " : "") << "__TJP::Result&)" << ResultBuffer::result_name();
       }
       code << ";" << endl;
     }
     
     if (tjp.target() || tjp.useAction() ||
-        (tjp.proceed () && ti->proceed_needs_target ())) {
+        (tjp.proceed () && proceed_needs_target (*loc))) {
       code << "  ";
       Naming::tjp_instance(code, loc);
       code << "._target = ";
-      if (ti->tjp_target () && WeaverBase::needs_this (ti->tjp_target())) {
+      if (target_func(loc) && needs_this (*target_func(loc))) {
         code << " (";
         if (tjp.useAction ())
           code << "void*)";
         else {
-          code << "__TJP::Target*)";
+          code << (is_dep ? "typename " : "") << "__TJP::Target*)";
         }
         switch (jptype) {
-#ifdef ACMODEL
         case JPT_Construction:
         case JPT_Destruction:
         case JPT_Execution:
           code << "this";
           break;
         case JPT_Call:
-#else
-        case JoinPointLoc::Construction:
-        case JoinPointLoc::Destruction:
-        case JoinPointLoc::Method:
-          code << "this";
-          break;
-        case JoinPointLoc::MethodCall:
-#endif
-          code << "dstthis";
+          code << "&dst";
             break;
           default:
             code << " 0";
@@ -518,33 +392,24 @@ void ThisJoinPoint::gen_tjp_init (ostream &code, JPL_Code *loc,
     }
 
     if (tjp.that() || tjp.useAction() ||
-        (tjp.proceed () && ti->proceed_needs_that ())) {
+        (tjp.proceed () && proceed_needs_that (*loc))) {
       code << "  ";
       Naming::tjp_instance(code, loc);
       code << "._that = ";
-      if (ti->tjp_that () && WeaverBase::needs_this (ti->tjp_that())) {
+      if (that_func (loc) && needs_this (*that_func (loc))) {
         code << " (";
         if (tjp.useAction ())
           code << "void*)";
         else {
-          code << "__TJP::That*)";
+          code << (is_dep ? "typename " : "")<< "__TJP::That*)";
         }
         switch (jptype) {
-#ifdef ACMODEL
         case JPT_Construction:
         case JPT_Destruction:
         case JPT_Execution:
           code << "this";
           break;
         case JPT_Call:
-#else
-        case JoinPointLoc::Construction:
-        case JoinPointLoc::Destruction:
-        case JoinPointLoc::Method:
-          code << "this";
-          break;
-        case JoinPointLoc::MethodCall:
-#endif
           code << "srcthis";
             break;
           default:
@@ -567,6 +432,28 @@ void ThisJoinPoint::gen_tjp_init (ostream &code, JPL_Code *loc,
 
 }
 
+void ThisJoinPoint::merge_flags (ACM_CodePlan &plan) {
+  if (plan.has_next_level())
+    merge_flags(*plan.get_next_level());
+  typedef ACM_Container<ACM_CodeAdvice, true> Container;
+  Container &before = plan.get_before();
+  for (Container::iterator i = before.begin(); i != before.end(); ++i)
+    TI_CodeAdvice::of(*(*i))->get_advice_info()->addTJPFlags(*this);
+  if (plan.has_around()) {
+    TI_CodeAdvice::of(*plan.get_around())->get_advice_info()->addTJPFlags(*this);
+    if (*TI_CodeAdvice::of(*plan.get_around())->get_condition())
+      conditional();
+  }
+  Container &after = plan.get_after();
+  for (Container::iterator i = after.begin(); i != after.end(); ++i)
+    TI_CodeAdvice::of(*(*i))->get_advice_info()->addTJPFlags(*this);
+}
+
+bool ThisJoinPoint::arg_needed (ACM_Code *loc) const {
+  return (arg() || useAction() || (proceed () &&
+    proceed_needs_args (*loc)));
+}
+
 void ThisJoinPoint::dump (ostream &out) const {
   out << "ThisJoinPoint (";
   out << "signature=" << (signature () ? "true" : "false");
