@@ -19,42 +19,41 @@
 #ifndef __transform_info_h__
 #define __transform_info_h__
 
-#ifdef ACMODEL
-#include "Elements.h"
-#else
-#include "JoinPointLoc.h"
-#endif
+#include "ACModel/Elements.h"
 #include "ThisJoinPoint.h"
+#include "PointCutExpr.h"
+#include "SyntacticContext.h"
+#include "CFlow.h"
+
 
 #include "Puma/CSemDatabase.h"
 #include "Puma/CNamespaceInfo.h"
 #include "Puma/DeducedArgument.h"
 #include "Puma/CTemplateInstance.h"
+#include "Puma/CFctInstance.h"
 #include "Puma/MacroUnit.h"
 
-class TransformInfo : public 
-#ifdef ACMODEL
-XModelTransformInfo
-#else
-ModelTransformInfo
-#endif
-{
+class TI_CodeAdvice : public ModelTransformInfo {
+  const Condition *_condition;
+  AdviceInfo *_advice_info;
+public:
+  void set_condition (const Condition *c) { _condition = c; }
+  const Condition *get_condition () const { return _condition; }
+  void set_advice_info (AdviceInfo *ai) { _advice_info = ai; }
+  AdviceInfo *get_advice_info () const { return _advice_info; }
+
+  static TI_CodeAdvice *of (const ACM_CodeAdvice &loc) {
+    return (TI_CodeAdvice*)loc.transform_info ();
+  }
+};
+
+class TransformInfo : public ModelTransformInfo {
 public:
-#ifdef ACMODEL
   virtual ModelNode &jpl () = 0;
-#else
-  virtual JoinPointLoc &jpl () = 0;
-#endif
   virtual CTree *tree () const = 0;
   virtual CObjectInfo *obj_info () const = 0;
   // associated unique parser object (e.g. for name mangling)
   virtual CObjectInfo *assoc_obj () const = 0;
-  int line () const {
-    return tree ()->token ()->location ().line ();
-  }
-  int lines () const {
-    return tree ()->end_token ()->location ().line () - line () + 1;
-  }
   // find the unit (non-macro) in which a join point is located
   Unit *unit () const {
     CTree *node = tree ();
@@ -68,25 +67,24 @@ public:
     return unit;
   }
 
-#ifdef ACMODEL
   static inline const TransformInfo *of (const ModelNode &loc);
   static inline CTree *tree (const ModelNode &loc);
-  static inline int line (const ModelNode &loc);
-  static inline int lines (const ModelNode &loc);
   static inline Unit *unit (const ModelNode &loc);
   static inline Location location (const ModelNode &loc);
-#else
-  static inline const TransformInfo *of (const JoinPointLoc &loc);
-  static inline CTree *tree (const JoinPointLoc &loc);
-  static inline int line (const JoinPointLoc &loc);
-  static inline int lines (const JoinPointLoc &loc);
-  static inline Unit *unit (const JoinPointLoc &loc);
-  static inline Location location (const JoinPointLoc &loc);
-#endif
-  
+
+  // helper function: check if a function is a method (needs a 'this' pointer)
+  static bool needs_this (CFunctionInfo *func) {
+    if (func->isMethod () && !func->isStaticMethod ()) {
+      return !(func->isOperator () &&
+        (strcmp (func->Name (), "operator new") == 0 ||
+         strcmp (func->Name (), "operator new[]") == 0 ||
+         strcmp (func->Name (), "operator delete") == 0 ||
+         strcmp (func->Name (), "operator delete[]") == 0));
+    }
+    return false;
+  }
 };
 
-#ifdef ACMODEL
 inline const TransformInfo *TransformInfo::of (const ModelNode &loc) {
   return loc.transform_info () ? (TransformInfo*)loc.transform_info () : 0;
 }
@@ -94,14 +92,6 @@ inline CTree *TransformInfo::tree (const ModelNode &loc) {
   return loc.transform_info () ?
     ((TransformInfo*)loc.transform_info ())->tree () : 0;
 }
-inline int TransformInfo::line (const ModelNode &loc) {
-  return loc.transform_info () ?
-    ((TransformInfo*)loc.transform_info ())->line () : -1;
-}
-inline int TransformInfo::lines (const ModelNode &loc) {
-  return loc.transform_info () ?
-    ((TransformInfo*)loc.transform_info ())->lines () : -1;
-}
 inline Unit *TransformInfo::unit (const ModelNode &loc) {
   return loc.transform_info () ?
     ((TransformInfo*)loc.transform_info ())->unit () : 0;
@@ -110,31 +100,6 @@ inline Location TransformInfo::location (const ModelNode &loc) {
   return loc.transform_info () ?
     tree (loc)->token ()->location () : Location ();
 }
-#else
-inline const TransformInfo *TransformInfo::of (const JoinPointLoc &loc) {
-  return loc.transform_info () ? (TransformInfo*)loc.transform_info () : 0;
-}
-inline CTree *TransformInfo::tree (const JoinPointLoc &loc) {
-  return loc.transform_info () ?
-    ((TransformInfo*)loc.transform_info ())->tree () : 0;
-}
-inline int TransformInfo::line (const JoinPointLoc &loc) {
-  return loc.transform_info () ?
-    ((TransformInfo*)loc.transform_info ())->line () : -1;
-}
-inline int TransformInfo::lines (const JoinPointLoc &loc) {
-  return loc.transform_info () ?
-    ((TransformInfo*)loc.transform_info ())->lines () : -1;
-}
-inline Unit *TransformInfo::unit (const JoinPointLoc &loc) {
-  return loc.transform_info () ?
-    ((TransformInfo*)loc.transform_info ())->unit () : 0;
-}
-inline Location TransformInfo::location (const JoinPointLoc &loc) {
-  return loc.transform_info () ?
-    tree (loc)->token ()->location () : Location ();
-}
-#endif
 
 class TI_Namespace : public TransformInfo {
   // pointer to the Puma namespace object (for transformation)
@@ -159,12 +124,23 @@ public:
 
 class TI_Class : public TransformInfo {
   CClassInfo *_class_obj;
-  int _phase;
-public:
-  TI_Class () : _class_obj (0), _phase (0) {}
 
-  void phase (int p) { _phase = p; }
-  int phase () const { return _phase; }
+  // internal function that checks whether a given scope is located within
+  // a template instance.
+  // TODO: this is a duplicate of a ModelBuilder function!
+  static bool is_in_template_instance (CScopeInfo *scope) {
+    if (scope->isTemplateInstance ())
+      return true;
+    if (scope->QualifiedScope ())
+      return is_in_template_instance (scope->QualifiedScope ());
+    if (scope->Parent () != scope)
+      return is_in_template_instance (scope->Parent ());
+    return false;
+  }
+
+public:
+  TI_Class () : _class_obj (0) {}
+
   void class_info (CClassInfo *c) { _class_obj = c->DefObject (); }
   CClassInfo *class_info () const { return _class_obj; }
 
@@ -172,53 +148,77 @@ public:
   virtual CObjectInfo *assoc_obj () const { return _class_obj; }
   virtual CTree *tree () const { return _class_obj->Tree (); }
   
+  // return the position of the opening bracket of the class body
+  Token *body_start_pos () const {
+    return ((CT_ClassDef*)_class_obj->Tree ())->Members ()->token ();
+  }
+
+  // return the position of the ";" after the class definition
+  Token *objdecl_end_pos () const {
+    CT_ClassDef *cd = (CT_ClassDef*)_class_obj->Tree();
+    return cd->ObjDecl ()->end_token ();
+  }
+
+  // return the position of the first token of the class definition
+  Token *objdecl_start_pos () const {
+    if (!_class_obj) return 0; // probably the *definition* is external; only phase 1 found a declaration
+    CT_ClassDef *cd = (CT_ClassDef*)_class_obj->Tree();
+    return cd->ObjDecl ()->token ();
+  }
+
+  // check whether this is a class or struct
+  bool is_class () const { return _class_obj->isClass(); }
+
+  // check whether this is a struct (more restrictive than 'is_class')
+  bool is_struct () const { return _class_obj->isStruct(); }
+
+  // check whether the class is defined (=has a body) and not only declared
+  bool is_defined () const { return _class_obj->isDefined(); }
+
+  // checks whether this class is a template *instance*
+  bool is_template_instance () const { return _class_obj->isTemplateInstance(); }
+
+  // check whether the class is defined in a extern "C" block
+  bool is_extern_c () const {
+    if (!is_defined())
+      return false;
+    CT_ClassDef *cd = (CT_ClassDef*)_class_obj->DefObject ()->Tree ();
+    // TODO: ugly workaround needed to check whether a class is defined in
+    // an extern "C" block -> extended Puma, SemDB doesn't contain this info!
+    for (CTree *curr = cd; curr; curr = curr->Parent())
+      if (curr->NodeName() == CT_LinkageSpec::NodeId())
+        return true;
+    return false;
+  }
+
+  // checks whther the class is defined within a template instance
+  bool is_in_template_instance () const {
+    return is_in_template_instance(_class_obj);
+  }
+
   static string name (CClassInfo *c) {
     ostringstream class_name;
     c->TypeInfo ()->TypeText (class_name, 0, false, false, false, true);
     return class_name.str ().c_str ();
   }
 
-#ifdef ACMODEL
-  static const TI_Class *of (const ACM_Class &loc) {
-#else
-  static const TI_Class *of (const JPL_Class &loc) {
-#endif
+  static TI_Class *of (const ACM_Class &loc) {
     return (TI_Class*)loc.transform_info ();
   }
 };
 
 class TI_Aspect : public TI_Class {
-  ACAspectInfo *_aspect_obj;
 public:
-  TI_Aspect () : _aspect_obj (0) {}
-    
-  void aspect_info (ACAspectInfo *ai) {
-    _aspect_obj = ai;
-    class_info (ai->ClassInfo ());
-  }
-  ACAspectInfo *aspect_info () const { return _aspect_obj; }
+
   CFunctionInfo *aspectof () const {
-    // make this better!
-    CFunctionInfo *aof = ClassInfo ()->Function ("aspectof");
+    // TODO: make this better!
+    CFunctionInfo *aof = class_info()->Function ("aspectof");
     if (!aof)
-      aof = ClassInfo ()->Function ("aspectOf");
+      aof = class_info()->Function ("aspectOf");
     return aof ? aof->DefObject () : (CFunctionInfo*)0;
   }
-  const char *name () const { return _aspect_obj->name (); }
-  ACAspectInfo *acnode () const { return _aspect_obj; }
-  CClassInfo *ClassInfo () const { return _aspect_obj->ClassInfo (); }
-  
-  static string name (ACAspectInfo *ai) {
-    ostringstream class_name;
-    ai->ClassInfo ()->TypeInfo ()->TypeText (class_name, 0, false, false, false, true);
-    return class_name.str ();
-  }
 
-#ifdef ACMODEL
   static const TI_Aspect *of (const ACM_Aspect &loc) {
-#else  
-  static const TI_Aspect *of (const JPL_Aspect &loc) {
-#endif
     return (TI_Aspect*)loc.transform_info ();
   }
 
@@ -226,29 +226,39 @@ public:
 
 class TI_Function : public TransformInfo {
   CFunctionInfo *_func_obj;
-  int _phase;
+  mutable vector<SyntacticContext> _contexts;
+
 public:
-  TI_Function () : _func_obj (0), _phase (0) {}
+  TI_Function () : _func_obj (0) {}
     
   void func_info (CFunctionInfo *fi) { _func_obj = fi->DefObject (); }
+  const vector<SyntacticContext> &syntactic_contexts () const {
+    if (_contexts.size() == 0) {
+      // Store a pointer to each declaration and the definition (if one exists)
+      CFunctionInfo *curr = _func_obj;
+      do {
+        _contexts.push_back (SyntacticContext (curr));
+        curr = curr->NextObject ()->FunctionInfo ();
+      }
+      while (curr != _func_obj);
+    }
+    return _contexts;
+  }
+
   CFunctionInfo *func_info () const { return _func_obj; }
-  void phase (int p) { _phase = p; }
-  int phase () const { return _phase; }
   virtual CObjectInfo *obj_info () const { return _func_obj; }
   virtual CObjectInfo *assoc_obj () const { return _func_obj; }
   virtual CTree *tree () const { return _func_obj->Tree (); }
 
   static string name (CFunctionInfo *func_info) {
-    ostringstream sig, name;
-    if (func_info->isMethod () && !func_info->isStaticMethod ())
-      name << func_info->Name ();
-    else
-      name << func_info->QualName (false, false, true);
+    ostringstream name;
+    // the unique name in a scope start with the function name
+    name << func_info->Name ();
       
     // add the template arguments if it is a function template instance
     CTemplateInstance *instance = func_info->TemplateInstance ();
     if (instance) {
-      name << "< ";
+      name << "<";
       for (unsigned a = 0; a < instance->DeducedArgs (); a++) {
         if (a > 0) name << ",";
         DeducedArgument *arg = instance->DeducedArg (a);
@@ -265,22 +275,55 @@ public:
         else
           name << "*invalid template arg*";
       }
-      name << " >";
+      string name_str = name.str ();
+      if (name_str[name_str.length () - 1] == '>')
+        name << " ";
+      name << ">";
     }
-    func_info->TypeInfo ()->TypeText (sig, name.str ().c_str (),
-        false, false, false, true);
-    return sig.str ();
+
+    // add the argument types
+    CTypeFunction *this_type = func_info->TypeInfo ();
+    name << "(";
+    CTypeList *argtypes = this_type->ArgTypes ();
+    for (unsigned i = 0; i < argtypes->Entries (); i++) {
+      if (i > 0) name << ",";
+      CTypeInfo *argtype = argtypes->Entry (i);
+      argtype->TypeText (name, "", false, false, false, true);
+    }
+    name << ")";
+
+    // add CV qualifiers
+    if (this_type->isConst ())
+      name << " const";
+    if (this_type->isVolatile ())
+      name << " volatile";
+
+    return name.str ();
   }
 
-#ifdef ACMODEL
   static const TI_Function *of (const ACM_Function &loc) {
-#else
-  static const TI_Function *of (const JPL_Function &loc) {
-#endif
     return (TI_Function*)loc.transform_info ();
   }
 };
 
+class TI_Variable : public TransformInfo {
+  CObjectInfo *_obj;
+public:
+  TI_Variable () : _obj (0) {}
+
+  void obj_info (CObjectInfo *oi) { _obj = oi->DefObject (); }
+  virtual CObjectInfo *obj_info () const { return _obj; }
+  virtual CObjectInfo *assoc_obj () const { return _obj; }
+  virtual CTree *tree () const { return _obj->Tree (); }
+public:
+  static string name (CObjectInfo *obj_info) {
+    return obj_info->Name ().c_str ();
+  }
+  static const TI_Variable *of (const ACM_Variable &loc) {
+    return (TI_Variable*)loc.transform_info ();
+  }
+};
+
 class TI_Type : public TransformInfo {
   CTypeInfo *_type_info;
 public:
@@ -295,33 +338,56 @@ public:
 
   static string name (CTypeInfo *type_info) {
     ostringstream out;
-#ifdef ACMODEL    
     type_info->TypeText(out,"?", false, false, false, true);
     string result = out.str();
     int last = result.length() - 1;
     if (result[last] == '?') last--;
     while(result[last] == ' ') last--;
     return result.substr(0, last + 1);
-#else
-    out << *type_info;
-    return out.str ();
-#endif
   }
-  
-#ifdef ACMODEL
+
   static const TI_Type *of (const ACM_Type &loc) {
-#else
-  static const TI_Type *of (const JPL_Type &loc) {
-#endif
     return (TI_Type*)loc.transform_info ();
   }
 };
 
-class TI_Code : public TransformInfo {
+class TI_Arg : public TransformInfo {
+  CTypeInfo *_type_info;
 public:
-  virtual CFunctionInfo *tjp_type() const = 0;
-  virtual CFunctionInfo *tjp_target() const = 0;
-  virtual CFunctionInfo *tjp_that() const = 0;
+  TI_Arg () : _type_info (0) {}
+
+  void type_info (CTypeInfo *ti) { _type_info = ti; }
+  CTypeInfo *type_info () const { return _type_info; }
+
+  virtual CObjectInfo *obj_info () const { return 0; }
+  virtual CObjectInfo *assoc_obj () const { return 0; }
+  virtual CTree *tree () const { return 0; }
+
+  static string signature (CTypeInfo *type_info, const string &name) {
+    ostringstream out;
+    type_info->TypeText(out, name.c_str(), false, false, false, true);
+    return out.str();
+  }
+
+  static const TI_Arg *of (const ACM_Arg &loc) {
+    return (TI_Arg*)loc.transform_info ();
+  }
+};
+
+class TI_Code : public TransformInfo {
+
+  CFlowList _triggers;
+
+public:
+
+  // consider a necessary cflow trigger at this join point
+  bool consider (const CFlow &cflow) {
+    _triggers.push_back (cflow);
+    return true;
+  }
+
+  // return the list of cflows that must be entered/left here
+  const CFlowList &cflows () const { return _triggers; }
 
   // that types (for the JoinPoint-API)
   virtual CTypeInfo *that_type () const = 0;
@@ -329,11 +395,6 @@ public:
   // target type (for the JoinPoint-API)
   virtual CTypeInfo *target_type () const = 0;
   
-  // interface needed to generate proper proceed code
-  virtual bool proceed_needs_that () const { return false; }
-  virtual bool proceed_needs_target () const { return false; }
-  virtual bool proceed_needs_fptr () const { return false; }
-  
   // helper functions for derived classes
   static CTypeInfo *get_that_type (CObjectInfo *obj) {
     CFunctionInfo *func = obj->FunctionInfo ();
@@ -353,17 +414,10 @@ public:
     return &CTYPE_VOID;
   }
 
-  // helper function: check if a function is a method (needs a 'this' pointer)
-  static bool needs_this (CFunctionInfo *func) {
-    if (func->isMethod () && !func->isStaticMethod ()) {
-      return !(func->isOperator () &&
-        (strcmp (func->Name (), "operator new") == 0 ||
-         strcmp (func->Name (), "operator new[]") == 0 ||
-         strcmp (func->Name (), "operator delete") == 0 ||
-         strcmp (func->Name (), "operator delete[]") == 0));
-    }
-    return false;
+  static TI_Code *of (const ACM_Code &loc) {
+    return (TI_Code*)loc.transform_info ();
   }
+
 };
 
 class TI_Method : public TI_Code {
@@ -376,16 +430,9 @@ public:
   virtual CObjectInfo *assoc_obj () const { return _func_info; }
   virtual CTree *tree () const { return _func_info->Tree (); }
 
-  virtual CFunctionInfo *tjp_type() const {return _func_info;}
-  virtual CFunctionInfo *tjp_target() const {return _func_info;}
-  virtual CFunctionInfo *tjp_that() const {return _func_info;}
-
   // that type (for the JoinPoint-API)
   virtual CTypeInfo *that_type () const { return get_that_type (_func_info); }
 
-  // interface needed to generate proper proceed code
-  virtual bool proceed_needs_that () const { return needs_this (_func_info); }
-  
   // target type (for the JoinPoint-API)
   virtual CTypeInfo *target_type () const {
     CRecord *scope = _func_info->ClassScope ();
@@ -397,27 +444,32 @@ class TI_MethodCall : public TI_Code {
   CFunctionInfo *called_func;
   CObjectInfo *caller_obj;
   CT_Call *node;
+  SyntacticContext _call_context;
+  SyntacticContext _dest_func_context;
+  CT_Expression *_target_expr;
+  bool _target_is_ptr;
 public:
-  TI_MethodCall () : called_func (0), caller_obj (0), node (0) {}
+  TI_MethodCall () : called_func (0), caller_obj (0), node (0),
+    _target_expr (0), _target_is_ptr (false) {}
 
-  void called (CFunctionInfo *c) { called_func = c; }
-  CFunctionInfo *called () const { return called_func; }
-  void caller (CObjectInfo *c) { caller_obj = c; }
+  void called (CFunctionInfo *c) { called_func = c; _dest_func_context.set_object(c); }
+//  CFunctionInfo *called () const { return called_func; }
+  void caller (CObjectInfo *c) { caller_obj = c; _call_context.set_object(c); }
   CFunctionInfo *caller () { 
     return caller_obj ? caller_obj->FunctionInfo () : 0;
   }
-  void tree (CT_Call *n) { node = n; }
+  void tree (CT_Call *n) { node = n; if (n) _target_expr = target_expr (_target_is_ptr); }
+  bool has_target_expr () const { return _target_expr; }
+  bool target_is_ptr () const { return _target_is_ptr; }
   virtual CTree *tree () const { return node; }
   virtual CObjectInfo *obj_info () const { return called_func; }
   virtual CObjectInfo *assoc_obj () const { return caller_obj; }
   
-  virtual CFunctionInfo *tjp_type() const { return called_func;}
-  virtual CFunctionInfo *tjp_target() const { return called_func;}
-  virtual CFunctionInfo *tjp_that() const { return caller_obj->FunctionInfo ();}
+  const SyntacticContext &call_context () const { return _call_context; }
+  const SyntacticContext &dest_func_context () const { return _dest_func_context; }
 
-  // interface needed to generate proper proceed code
-  virtual bool proceed_needs_target () const { return needs_this (called_func); }
-  virtual bool proceed_needs_fptr () const { return needs_rights (); }
+  Token *before_pos () { return CallNode()->token(); }
+  Token *after_pos () { return CallNode()->end_token(); }
 
   // the result type
   CTypeInfo *result_type_info () const {
@@ -462,6 +514,10 @@ public:
       return &CTYPE_VOID;
   }
 
+  bool is_call_op () const {
+    return strcmp (called_func->Name().c_str (), "operator ()") == 0;
+  }
+
   // the target object of the call or NULL
   CT_Expression *target_expr (bool &is_ptr) const {
   
@@ -479,6 +535,11 @@ public:
   
     if (calltype == CT_CallExpr::NodeId ()) {
       CTree *expr = ((CT_CallExpr*)node)->Expr ();
+      // if the target function is given in parenthesis, skip it:
+      // e.g. (foo->bar)(42);
+      while (expr->NodeName () == CT_BracedExpr::NodeId ())
+        expr = ((CT_BracedExpr*)expr)->Expr ();
+      // find the target object if there is any
       const char *fctcalltype = expr->NodeName ();
       if (fctcalltype == CT_MembRefExpr::NodeId ()) {
         // <target>.method()
@@ -488,15 +549,22 @@ public:
         is_ptr = true;
         result = expr->Son (0);
       } else {
+        if (is_call_op ()) {
+          // <target>()
+          result = expr;
+        }
+        else {
         // method()
         is_ptr = true; // here 'this' is passed implicitly
+        }
       }
     }
     else if (calltype == CT_BinaryExpr::NodeId ()) {
       // <target> <op> <arg>
       result = node->Son (0);
     }
-    else if (calltype == CT_UnaryExpr::NodeId ()) {
+    else if (calltype == CT_UnaryExpr::NodeId () ||
+              calltype == CT_DerefExpr::NodeId ()) {
       // <op> <target>
       result = node->Son (1);
     }
@@ -537,6 +605,21 @@ public:
            nodename == CT_RootQualName::NodeId ();
   }
   
+  // checks whether the call uses explicit template parameters
+  bool has_explicit_template_params () const {
+    if (node->NodeName () != CT_CallExpr::NodeId ())
+      return false;
+    CTree *expr = ((CT_CallExpr*)node)->Expr ();
+    const char *nodename = expr->NodeName ();
+    if (nodename == CT_MembPtrExpr::NodeId () ||
+        nodename == CT_MembRefExpr::NodeId ()) {
+      expr = expr->Son (2); // some access function in PUMA missing!
+      nodename = expr->NodeName ();
+    }
+    return expr->IsSimpleName() &&
+        expr->IsSimpleName()->Name()->NodeName() == CT_TemplateName::NodeId();
+  }
+
   // returns true if the call needs special access rights
   bool needs_rights () const {
     // get the target object type
@@ -569,10 +652,6 @@ public:
   virtual CObjectInfo *assoc_obj () const { return _func_info; }
   virtual CTree *tree () const { return _func_info->ClassScope ()->Tree (); }
   
-  virtual CFunctionInfo *tjp_type() const {return _func_info;}
-  virtual CFunctionInfo *tjp_target() const {return _func_info;}
-  virtual CFunctionInfo *tjp_that() const {return _func_info;}
-
   // that type (for the JoinPoint-API)
   virtual CTypeInfo *that_type () const { return get_that_type (_func_info); }
 
@@ -581,9 +660,6 @@ public:
     CRecord *scope = _func_info->ClassScope ();
     return scope ? scope->TypeInfo () : &CTYPE_VOID;
   }
-
-  // interface needed to generate proper proceed code
-  virtual bool proceed_needs_that () const { return true; }
 };
 
 class TI_Destruction : public TI_Code {
@@ -596,10 +672,6 @@ public:
   virtual CObjectInfo *assoc_obj () const { return _func_info; }
   virtual CTree *tree () const { return _func_info->ClassScope ()->Tree (); }
 
-  virtual CFunctionInfo *tjp_type() const {return _func_info;}
-  virtual CFunctionInfo *tjp_target() const {return _func_info;}
-  virtual CFunctionInfo *tjp_that() const {return _func_info;}
-
   // that type (for the JoinPoint-API)
   virtual CTypeInfo *that_type () const { return get_that_type (_func_info); }
 
@@ -608,93 +680,53 @@ public:
     CRecord *scope = _func_info->ClassScope ();
     return scope ? scope->TypeInfo () : &CTYPE_VOID;
   }
-
-  // interface needed to generate proper proceed code
-  virtual bool proceed_needs_that () const { return true; }
 };
 
 class TI_AdviceCode : public TransformInfo {
-  CT_AdviceDecl *_tree;
+  // CT_AdviceDecl *_tree;
+  CFunctionInfo *_func;
   ThisJoinPoint _this_join_point;
-  int _phase;
   
 public:
-  TI_AdviceCode () : _tree (0), _phase (0) {}
+  TI_AdviceCode () : _func (0) {}
   
-  void phase (int p) { _phase = p; }
-  int phase () const { return _phase; }
-  void tree (CT_AdviceDecl *ad) { _tree = ad; }
-  virtual CTree *tree () const { return _tree; }
+  void function(CFunctionInfo *f) { _func = f; }
+  virtual CTree *tree () const { return _func ? _func->Tree() : 0; }
   
   virtual CObjectInfo *obj_info () const {
-    return ((CT_FctDef*)_tree->Decl ())->Object ();
+    return _func;
   }
   virtual CObjectInfo *assoc_obj () const { return obj_info (); }
 
   CScopeInfo *Scope () const {
-    return ((CT_FctDef*)_tree->Decl ())->Object ()->QualifiedScope ();
+    return _func ? _func->CObjectInfo::QualifiedScope () : 0;
   }
   const char *name () const {
-    return ((CT_FctDef*)_tree->Decl ())->Object ()->Name ();
+    return _func ? _func->Name () : 0;
   }
+  string qual_name () {
+    return _func ? _func->QualName () : "";
+  }
+  
   CFunctionInfo *function() const {
-    stringstream fname;
-    fname << "__" << (name () + 1);
-    return Scope ()->ClassInfo ()->Function (fname.str ().c_str ());
+    return _func;
   }
 
   ThisJoinPoint &this_join_point () { return _this_join_point; }
   const ThisJoinPoint &this_join_point () const { return _this_join_point; }
-  
-  static string name (CT_AdviceDecl *ad) {
-    const char *name = ((CT_FctDef*)ad->Decl ())->Object ()->Name () + 1;
-    CScopeInfo *scope = ((CT_FctDef*)ad->Decl ())->Object ()->QualifiedScope ();
-    
-    stringstream fname;
-    fname << scope->QualName () << "::__" << name;
-    return fname.str ();
-  }
-  
-#ifdef ACMODEL
+
   static TI_AdviceCode *of (const ACM_AdviceCode &loc) {
-#else
-  static TI_AdviceCode *of (const JPL_AdviceCode &loc) {
-#endif
     return (TI_AdviceCode*)loc.transform_info ();
   }
 };
 
 class TI_Introduction : public TransformInfo {
-  ACIntroductionInfo *_acii;
-  // the following elements are only needed for the *old* repository code
-  Location _loc;
-  int _lines;
-  Unit *_unit;
-  
 public:
-  TI_Introduction () : _acii (0) {}
-  
-  void intro_info (ACIntroductionInfo *acii) {
-    _acii = acii;
-    _loc = tree ()->token ()->location ();
-    _lines = tree ()->end_token ()->location ().line () - line () + 1;
-    _unit = (Unit*)tree ()->token ()->belonging_to ();
-  }
-      
   virtual CObjectInfo *obj_info () const { return 0; }
-  virtual CObjectInfo *assoc_obj () const { return obj_info (); }
-  virtual CT_AdviceDecl *tree () const { return _acii->def_node (); }
+  virtual CObjectInfo *assoc_obj () const { return 0; }
+  virtual CT_AdviceDecl *tree () const { return 0; }
 
-  int line () const { return _loc.line (); }
-  int lines () const { return _lines; }
-  Unit *unit () const { return _unit; }
-  Location loc () const { return _loc; }
-
-#ifdef ACMODEL
   static TI_Introduction *of (const ACM_Introduction &loc) {
-#else
-  static TI_Introduction *of (const JPL_Introduction &loc) {
-#endif
     return (TI_Introduction*)loc.transform_info ();
   }
 };
@@ -710,40 +742,89 @@ public:
   virtual CObjectInfo *assoc_obj () const { return obj_info (); }
   virtual CTree *tree () const { return _tree; }
 
-#ifdef ACMODEL
   static TI_Order *of (const ACM_Order &loc) {
-#else
-  static TI_Order *of (const JPL_Order &loc) {
-#endif
     return (TI_Order*)loc.transform_info ();
   }
 };
 
-#ifdef ACMODEL
-class TI_PointcutExpr : public TransformInfo {
-  CTree *_tree;
+class TI_Pointcut : public TransformInfo {
+  CFunctionInfo *_func_obj;
+  PointCutExpr *_pce;
+  Location _loc;
 public:
-  TI_PointcutExpr () : _tree (0) {}
-  void tree (CTree *node) { _tree = node; }
-  virtual CObjectInfo *obj_info () const { return 0; }
-  virtual CObjectInfo *assoc_obj () const { return 0; }
-  virtual CTree *tree () const { return _tree; }
+  TI_Pointcut () : _func_obj(0), _pce (0) {}
+  ~TI_Pointcut () { PointCutExpr::destroy(_pce); }
 
-  static TI_PointcutExpr *of (const ACM_PointcutExpr &loc) {
-    return (TI_PointcutExpr*)loc.transform_info ();
+  void func_info (CFunctionInfo *fi) { _func_obj = fi->DefObject (); }
+  CFunctionInfo *func_info () const { return _func_obj; }
+  virtual CObjectInfo *obj_info () const { return _func_obj; }
+  virtual CObjectInfo *assoc_obj () const { return _func_obj; }
+  virtual CTree *tree () const { return _func_obj->Tree (); }
+
+  void set_pce (PointCutExpr *pce) { _pce = pce; }
+  PointCutExpr *get_pce () const { return _pce; }
+  void set_location (const Location &loc) { _loc = loc; }
+  const Location &get_location () const { return _loc; }
+
+  static TI_Pointcut *of (const ACM_Pointcut &loc) {
+    return (TI_Pointcut*)loc.transform_info ();
   }
 };
-#endif
 
 class TI_ClassSlice : public TransformInfo {
   CObjectInfo *_obj;
   const Unit *_slice_unit;
-
-#ifdef ACMODEL
   Unit _pattern;
-#endif
+
+  // new phase 1 implementation:
+  Unit *_tokens;                             // class slice body
+  std::list<Unit*> _non_inline_members;      // members defined outside of body
+  std::list<Unit*> _non_inline_member_units; // corresponding source units
+
 public:
-  TI_ClassSlice () : _obj (0), _slice_unit (0) {}
+  TI_ClassSlice () : _obj (0), _slice_unit (0), _tokens (0) {}
+
+  // new phase 1 implementation:
+  ~TI_ClassSlice () {
+    delete _tokens;
+    for (std::list<Unit*>::const_iterator i = _non_inline_members.begin ();
+        i != _non_inline_members.end (); ++i)
+      delete *i;
+  }
+  void set_tokens (Unit *unit) { _tokens = unit; }
+  Unit *get_tokens () const { return _tokens; }
+  std::list<Unit*> &non_inline_members () { return _non_inline_members; }
+  std::list<Unit*> &non_inline_member_units () { return _non_inline_member_units; }
+  void analyze_tokens (bool &has_base_intro, bool &has_member_intro) {
+    has_base_intro = false;
+    has_member_intro = false;
+    Token *token = _tokens->first();
+    while (token && token->type () != TOK_SEMI_COLON) {
+      if (token->type () == TOK_COLON)
+        has_base_intro = true;
+      else if (token->type () == TOK_OPEN_CURLY) {
+        has_member_intro = true;
+        break;
+      }
+      token = _tokens->next(token);
+    }
+  }
+  std::string base_intro () const {
+    std::string result = "";
+    Token *token = _tokens->first();
+    token = _tokens->next(token);
+    while (token && token->type () != TOK_COLON)
+      token = _tokens->next(token);
+    token = _tokens->next(token); // skip ':'
+    while (token && token->type () != TOK_SEMI_COLON &&
+        token->type () != TOK_OPEN_CURLY) {
+      result += " ";
+      result += token->text ();
+      token = _tokens->next(token);
+    }
+    return result;
+  }
+  // end - new phase 1 implementation
 
   void obj_info (CObjectInfo *obj) { _obj = obj; }
   virtual CObjectInfo *obj_info () const { return _obj; }
@@ -755,13 +836,9 @@ public:
   virtual CTree *tree () const { return _obj->Tree (); }
 
 
-#ifdef ACMODEL
   const Unit &pattern () const { return _pattern; }
   Unit &pattern () { return _pattern; }
   static TI_ClassSlice *of (const ACM_ClassSlice &loc) {
-#else
-  static TI_ClassSlice *of (const JPL_ClassSlice &loc) {
-#endif
     return (TI_ClassSlice*)loc.transform_info ();
   }
   
