@@ -265,18 +265,33 @@ bool PreMacroExpander::checkForGoOn (bool mode) const
 // Build a unit from a string.
 Unit* PreMacroExpander::buildUnit (const char* string) const
  {
-    int i;
-
-    // Skip leading white spaces.
-    for (i = 0; StrCol::isSpace (string[i]); i++);
-    string = &string[i];
-
     // Create the component scanner and an empty unit.
     Unit *unit = new MacroUnit ();
 
-    // Scan the string and fill the unit.
+    // Allow macro concatenation and stringification operators
     _parser->cscanner ().allow_macro_ops ();
-    _parser->cscanner ().fill_unit (string, *unit);
+
+    // Skip leading white spaces.
+    while (StrCol::isSpace(*string)) string++;
+
+    int len = strlen(string);
+    if (StrCol::isSpace(string[len-1]))
+    {
+        // Skip trailing white spaces.
+        while (len > 1 && StrCol::isSpace(string[len-1])) len--;
+        char* buf = StrCol::dup (string, len);
+
+        // Scan the string and fill the unit.
+        _parser->cscanner ().fill_unit (buf, *unit);
+        delete[] buf;
+    }
+    else
+    {
+        // Scan the string and fill the unit.
+        _parser->cscanner ().fill_unit (string, *unit);
+    }
+
+    // Reset scanner
     _parser->cscanner ().allow_macro_ops (false);
 
     // Return the filled unit.
@@ -308,16 +323,20 @@ bool PreMacroExpander::equal (const char* str1, const char* str2) const
 // Test whether a string is an argument of a macro.
 int PreMacroExpander::isArg (const char* string, PreMacro* macro) const
  {
-    PreMacro::MacroArgs *args = macro->getArgs ();
+    const PreMacro::MacroArgs *args = macro->getArgs ();
+    if (!args) return -1; // -> no args, no hit
 
     // Go through the argument list of the macro and compare the
     // string with the arguments.
-    for (int i = 0; i < macro->numArgs (); i++)
+    int i = 0;
+    for (PreMacro::MacroArgs::const_iterator iter = args->begin ();
+        iter != args->end (); ++iter)
     {
         // Return the index of the argument in argument array
         // if found.
-        if (equal (string, args->lookup (i)))
+        if (equal (string, (*iter).c_str ()))
             return i;
+        i++;
     }
 
     // -1 signals that it`s not an argument.
@@ -474,8 +493,8 @@ void PreMacroExpander::replaceStringifiedArg (int index, Unit* unit, PreArgDesc
             for (token = (Token*) argUnit->first (); token;
                  token = (Token*) argUnit->next (token))
             {
-                // Escape double quotes and backslashes in strings.
-                if (token->type () == TOK_STRING_VAL)
+                // Escape double quotes and backslashes in strings or character constants
+                if (token->type () == TOK_STRING_VAL || token->type () == TOK_CHAR_VAL)
                 {
                     const char *strval = (const char*) token->text ();
                     unsigned int len = (unsigned int) strlen (strval);
@@ -520,7 +539,7 @@ void PreMacroExpander::processStringOp (PreMacro* macro, Unit* unit, PreArgDesc
     // Scan the unit.
     for (; token; token = (Token*) unit->next (token))
     {
-        // Recognize the stringification operator `#'.
+      // Recognize the stringification operator `#'.
         if (token->is_macro_op () && token->type () == TOK_MO_HASH)
         {
             if (! hash) hash = token;          // Remember a single hashmark.
@@ -624,7 +643,6 @@ void PreMacroExpander::processConcatOp (Unit* unit) const {
       next = tmp.first ();
       for (curr = next; curr; curr = tmp.next (curr)) {
         curr->location (op->location ());
-        curr->macro_generated ();
       }
 
       // move the generated tokens behind the expression <x> ## <y>
@@ -662,11 +680,10 @@ void PreMacroExpander::substituteMacro (PreMacro* macro, PreArgDesc &args) const
         processStringOp (macro, unit, args);
 
     // Set the location of every token in unit to the location of the
-    // macro call. Mark every token as macro generated.
+    // macro call.
     for (Token* token = (Token*) unit->first (); token;
          token = (Token*) unit->next (token)) {
         token->location (args.beginToken ()->location ());
-        token->macro_generated ();
     }
 
     // Process the concatenation macro operator `##'.
